#!/usr/bin/env python

""" Provides outliers removal methods.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@yale.edu>, Yale University
"""

import ctypes
from dataclasses import dataclass, field
import enum
import logging
import numpy as np
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

from pachyderm import histogram
from pachyderm import projectors
from pachyderm.typing_helpers import Hist
from pachyderm import utils

logger = logging.getLogger(__name__)

# Typing helper
OutliersRemovalAxis = Union[projectors.TH1AxisType, enum.Enum]

def _get_mean_and_median(hist: Hist) -> Tuple[float, float]:
    """ Retrieve the mean and median from a ROOT histogram.

    Note:
        These values are not so trivial to calculate without ROOT, as they are the bin values
        weighted by the bin content.

    Args:
        hist: Histogram from which the values will be extract.
    Returns:
        mean, median of the histogram.
    """
    # Median
    # See: https://root-forum.cern.ch/t/median-of-histogram/7626/5
    x = ctypes.c_double(0)
    q = ctypes.c_double(0.5)
    # Apparently needed to be safe(?)
    hist.ComputeIntegral()
    hist.GetQuantiles(1, x, q)

    mean = hist.GetMean()

    return (mean, x.value)

@dataclass
class _OutputObject:
    """ Helper object to retrieve the result of a projector. """
    output: Hist

def _project_to_part_level(hist: Hist, outliers_removal_axis: OutliersRemovalAxis) -> Hist:
    """ Project the input histogram to the particle level axis.

    Args:
        hist: Histogram to check for outliers.
        outliers_removal_axis: Axis along which outliers removal will be performed. Usually
            the particle level aixs.
    Returns:
        The histogram to check for outliers.
    """
    # Setup the projector
    import ROOT
    if isinstance(hist, (ROOT.TH2, ROOT.TH3)):
        projection_information: Dict[str, Any] = {}
        output_object = _OutputObject(None)
        projector = projectors.HistProjector(
            observable_to_project_from = hist,
            output_observable = output_object,
            output_attribute_name = "output",
            projection_name_format = "outliers_removal_hist",
            projection_information = projection_information,
        )
        # No additional_axis_cuts or projection_dependent_cut_axes
        # Projection axis
        projector.projection_axes.append(
            projectors.HistAxisRange(
                axis_type = outliers_removal_axis,
                axis_range_name = "outliers_removal_axis",
                min_val = projectors.HistAxisRange.apply_func_to_find_bin(None, 1),
                max_val = projectors.HistAxisRange.apply_func_to_find_bin(ROOT.TAxis.GetNbins),
            )
        )

        # Perform the actual projection and return the output.
        projector.project()
        return output_object.output

    # If we already have a 1D hist, just return that existing hist.
    return hist

def _determine_outliers_index(hist: Hist,
                              moving_average_threshold: float = 1.0,
                              number_of_values_to_search_ahead: int = 5,
                              limit_of_number_of_values_below_threshold: Optional[int] = None) -> int:
    """ Determine the location of where outliers begin in a 1D histogram.

    When the moving average falls below the limit, we consider the outliers to have begun.

    To determine the location of outliers:

    - Calculate the moving average for number_of_values_to_search_ahead values.
    - First, the moving average must go above the limit at least once to guard against a random cut
      in a low pt bin causing most of the data to be cut out.
    - Next, we look for a consecutive number of entries below limit_of_number_of_values_below_threshold.
    - If we meet that condition, we have found the index where the outliers begin. We then return the ROOT
      bin index of the value.
    - If not, we return -1.

    Note:
        The index returned is when the moving average first drops below the threshold for a moving average
        calculated with that bin at the center. This is somewhat different from a standard moving average
        calculation which would only look forward in the array.

    Args:
        hist: Histogram to be checked for outliers.
        moving_average_threshold: Value of moving average under which we consider the moving average
            to be 0. Default: 2.
        number_of_values_to_search_ahead: Number of values to search ahead in the array when calculating
            the moving average. Default: 5.
        limit_of_number_of_values_below_threshold: Number of consecutive bins below the threshold to be considered
            the beginning of outliers. Default: None, which will correspond to number_of_values_to_search_ahead - 1.
    Returns:
        ROOT (ie 1-indexed) index of the histogram axes where the outliers begin.
    """
    # Validation
    import ROOT
    if isinstance(hist, (ROOT.TH2, ROOT.TH3, ROOT.THnBase)):
        raise ValueError(
            f"Given histogram '{hist.GetName()}' of type {type(hist)}, but can only"
            " determine the outlier location of a 1D histogram. Please project to"
            " the particle level axis first."
        )

    if limit_of_number_of_values_below_threshold is None:
        # In principle, this could be another value. However, this is what was used in the previous outliers
        # removal implementation.
        limit_of_number_of_values_below_threshold = number_of_values_to_search_ahead - 1

    # It is much more convenient to work with a numpy array.
    hist_to_check = histogram.Histogram1D.from_existing_hist(hist)

    # Calculate the moving average for the entire axis, looking ahead including the current bin + 4 = 5 ahead.
    number_of_values_to_search_ahead = 5
    moving_average = utils.moving_average(hist_to_check.y, n = number_of_values_to_search_ahead)

    #logger.debug(f"y: {hist_to_check.y}")
    #logger.debug(f"moving_average: {moving_average}")

    cut_index = _determine_outliers_for_moving_average(
        moving_average = moving_average,
        moving_average_threshold = moving_average_threshold,
        number_of_values_to_search_ahead = number_of_values_to_search_ahead,
        limit_of_number_of_values_below_threshold = limit_of_number_of_values_below_threshold,
    )

    if cut_index != -1:
        # ROOT histograms are 1 indexed, so we add another 1.
        cut_index += 1

    return cut_index

def _determine_outliers_for_moving_average(moving_average: np.ndarray,
                                           moving_average_threshold: float,
                                           number_of_values_to_search_ahead: int,
                                           limit_of_number_of_values_below_threshold: int) -> int:
    """ Determine outliers to remove from a given moving average.

    Note:
        The index returned is when the moving average first drops below the threshold for a moving average
        calculated with that bin at the center. This is somewhat different from a standard moving average
        calculation which would only look forward in the array.

    Args:
        moving_average: Moving average.
        moving_average_threshold: Value of moving average under which we consider the moving average
            to be 0. Default: 2.
        number_of_values_to_search_ahead: Number of values to search ahead in the array when calculating
            the moving average. Default: 5.
        limit_of_number_of_values_below_threshold: Number of consecutive bins below the threshold to be considered
            the beginning of outliers. Default: None, which will correspond to number_of_values_to_search_ahead - 1.
    Returns:
        0-indexed index of the histogram axes where the outliers begin.
    """
    below_threshold = moving_average < moving_average_threshold

    # Build up a list of values to check if they are below threshold. This list allows us to easily look
    # forward in the below_threshold array.
    values_to_check = []
    for i in range(limit_of_number_of_values_below_threshold):
        # Basically, this gives us (for limit_of_number_of_values_below_threshold = 4):
        # below_threshold[0:-3], below_threshold[1:-2], below_threshold[2:-1], below_threshold[3:None]
        values_to_check.append(
            below_threshold[i:-(limit_of_number_of_values_below_threshold - 1 - i) or None]
        )

    # Some helpful logging information.
    #logger.debug(f"values_to_check: {values_to_check}")
    #logger.debug(f"moving avg length: {len(moving_average)}, length of values_to_check entries: {[len(v) for v in values_to_check]}")

    # Must have at least one bin above the specified threshold.
    found_at_least_one_bin_above_threshold = False
    # Index we will search for from which outliers will be cut.
    cut_index = -1

    # Determine the index where the limit_of_number_of_values_below_threshold bins are consequentially below the threshold.
    for i, values in enumerate(zip(*values_to_check)):
        # Skip the first bin because some old pt hard bin trains had a large number of erroneous entries
        # in the first bin (regardless of the actual pt hard bin). This should be resolved in the embedding
        # helper now. In any case, it doesn't make sense to encounter outliers in the first bin, so this is a
        # fine bin to skip.
        if i == 0:
            continue

        # True if below threshold, so check if not True.
        above_threshold = [not value for value in values]
        # We require the values to go above the moving average threshold at least once.
        if any(above_threshold):
            #logger.debug(f"Found bin i {i} above threshold with moving average: {moving_average[i]}")
            found_at_least_one_bin_above_threshold = True

        # All values from which we are looking ahead must be below the threshold to consider the index
        # as below threshold.
        if found_at_least_one_bin_above_threshold and all(np.invert(above_threshold)):
            # The previous outlier removal implementation used a moving average centered on a value
            # (ie. it checked ``arr[-2 + current_index:current_index + 3]``). Thus, we need to
            # shift the cut_index that we assign by limit_of_number_of_values_below_threshold // 2 for
            # the index where we have found all values below the threshold.
            logger.debug(f"i at found cut_index: {i} with moving_average: {moving_average[i]}")
            cut_index = i + limit_of_number_of_values_below_threshold // 2
            break

    return cut_index

def _remove_outliers_from_hist(hist: Hist, outliers_start_index: int, outliers_removal_axis: OutliersRemovalAxis) -> None:
    """ Remove outliers from a given histogram.

    Args:
        hist: Histogram to check for outliers.
        outliers_start_index: Index in the truth axis where outliers begin.
        outliers_removal_axis: Axis along which outliers removal will be performed. Usually
            the particle level aixs.
    Returns:
        None. The histogram is modified in place.
    """
    # Use on TH1, TH2, and TH3 since we don't start removing immediately, but instead only after the limit
    if outliers_start_index > 0:
        #logger.debug("Removing outliers")
        # Check for values above which they should be removed by translating the global index
        x = ctypes.c_int(0)
        y = ctypes.c_int(0)
        z = ctypes.c_int(0)
        # Maps axis to valaues
        # This is kind of dumb, but it works.
        outliers_removal_axis_values: Dict[OutliersRemovalAxis, ctypes.c_int] = {
            projectors.TH1AxisType.x_axis: x,
            projectors.TH1AxisType.y_axis: y,
            projectors.TH1AxisType.z_axis: z,
        }
        for index in range(0, hist.GetNcells()):
            # Get the bin x, y, z from the global bin
            hist.GetBinXYZ(index, x, y, z)
            # Watch out for any problems
            if hist.GetBinContent(index) < hist.GetBinError(index):
                logger.warning(f"Bin content < error. Name: {hist.GetName()}, Bin content: {hist.GetBinContent(index)}, Bin error: {hist.GetBinError(index)}, index: {index}, ({x.value}, {y.value})")
            if outliers_removal_axis_values[outliers_removal_axis].value >= outliers_start_index:
                #logger.debug("Cutting for index {}. x bin {}. Cut index: {}".format(index, x, cutIndex))
                hist.SetBinContent(index, 0)
                hist.SetBinError(index, 0)
    else:
        logger.info(f"Hist {hist.GetName()} did not have any outliers to cut")

@dataclass
class OutliersRemovalManager:
    moving_average_threshold: float = field(default = 1.0)

    def run(self, outliers_removal_axis: OutliersRemovalAxis,
            hist: Optional[Hist] = None, hists: Optional[Mapping[str, Hist]] = None,
            mean_fractional_difference_limit: float = 0.01,
            median_fractional_difference_limit: float = 0.01,) -> int:
        """ Remove outliers from the given histogram(s).

        Args:
            outliers_removal_axis: Axis along which outliers removal will be performed. Usually
                the particle level aixs.
            hist: Histogram to check for outliers. Either this or ``hists`` must be specified.
            hists: Histograms to check for outliers. Either this or ``hist`` must be specified.
            mean_fractional_difference_limit: Max fractional difference of mean after outliers
                removal. Default: 0.01.
            median_fractional_difference_limit: Max fractional difference of median after outliers
                removal. Default: 0.01.
        Return:
            Bin index value from which the outliers were removed. The histogram(s) is modified in place.
        """
        # Validation
        if hist is None and hists is None:
            raise ValueError("Must specify either a single hist or a sequence of hists.")
        if hist and hists:
            raise ValueError("Cannot specify both a single hist and a sequence of hists.")
        # Convert the hist into a temporary list so that we can use the same code below.
        if hist is not None:
            hists = {"hist": hist}
        # To help mypy typing
        assert hists is not None
        # Final hist validation
        for h in hists:
            if hasattr(h, "ProjectionND") and hasattr(h, "Projection"):
                raise ValueError("Cannot remove outliers from THn hists. Project to TH3 or lower first.")

        # Keep track of the outliers index for each hist to determine the maximum of the hists
        # that are passed in.
        outliers_indices: List[int] = []
        # Keep track of pre/post median values
        pre_removal_mean = {}
        post_removal_mean = {}
        pre_removal_median = {}
        post_removal_median = {}

        for hist_name, hist in hists.items():
            # Setup
            hist_to_check = _project_to_part_level(hist = hist, outliers_removal_axis = outliers_removal_axis)

            # Check these values before and after outlier removal.
            (pre_removal_mean[hist_name], pre_removal_median[hist_name]) = _get_mean_and_median(hist_to_check)

            # Determine the index where the outliers begin and then remove them.
            outliers_indices.append(
                _determine_outliers_index(
                    hist = hist_to_check,
                    moving_average_threshold = self.moving_average_threshold,
                )
            )

        outliers_start_index: int = np.max(outliers_indices)
        logger.debug(f"outliers_start_index: {outliers_start_index}")

        for hist_name, hist in hists.items():
            # Then do the actual outliers removal.
            _remove_outliers_from_hist(
                hist = hist,
                outliers_start_index = outliers_start_index,
                outliers_removal_axis = outliers_removal_axis,
            )

            # Now check the mean and median to see how much they've changed.
            hist_to_check = _project_to_part_level(hist = hist, outliers_removal_axis = outliers_removal_axis)
            (post_removal_mean[hist_name], post_removal_median[hist_name]) = _get_mean_and_median(hist_to_check)
            mean_fractional_difference = (post_removal_mean[hist_name] - pre_removal_mean[hist_name]) / post_removal_mean[hist_name]
            median_fractional_difference = (post_removal_median[hist_name] - pre_removal_median[hist_name]) / post_removal_median[hist_name]
            logger.info(
                f"Hist {hist_name}: pre- vs post- outliers removal mean fractional difference:"
                f" {mean_fractional_difference}, median fractional difference: {median_fractional_difference}"
            )
            # Provide some very broad checks. We should be very surprised if the post outliers values vary by more than 5%
            if mean_fractional_difference > mean_fractional_difference_limit:
                raise RuntimeError(
                    f"Mean fractional difference is greater than {mean_fractional_difference_limit * 100}%!"
                    f" {mean_fractional_difference * 100}%"
                )
            if median_fractional_difference > median_fractional_difference_limit:
                raise RuntimeError(
                    f"Median fractional difference is greater than {median_fractional_difference_limit * 100}%!"
                    f" {median_fractional_difference * 100}%"
                )

        logger.debug(f"Outliers removal complete! Found outliers_start_index: {outliers_start_index}")
        return outliers_start_index




#!/usr/bin/env python

""" Plotting styling and utilities.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import matplotlib

def restore_defaults() -> None:
    """ Restore the default matplotlib settings. """
    matplotlib.rcParams.update(matplotlib.rcParamsDefault)

def configure() -> None:
    """ Configure matplotlib according to my (biased) specification.

    As a high level summary, this is a combination of a number of seaborn settings, along with
    my own tweaks. By calling this function, the matplotlilb ``rcParams`` will be modified according
    to these settings.

    Up to this point, the settings have been configured by importing the `jet_hadron.plot.base`
    module, which set a variety of parameters on import. This included some options which were set
    by seaborn. Additional modifications were made to the fonts to ensure that they are the same in
    labels and latex. Lastly, it tweaked smaller visual settings. The differences between the default
    matplotlib and these settings are:

    .. code-block:: python
        >>> pprint.pprint(diff)
        {'axes.axisbelow': 'original: line, new: True',
         'axes.edgecolor': 'original: black, new: .15',
         'axes.labelcolor': 'original: black, new: .15',
         'axes.labelsize': 'original: medium, new: 12.0',
         'axes.linewidth': 'original: 0.8, new: 1.25',
         'axes.prop_cycle': "original: cycler('color', ['#1f77b4', '#ff7f0e', "
                            "'#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', "
                            "'#7f7f7f', '#bcbd22', '#17becf']), new: cycler('color', "
                            '[(0.2980392156862745, 0.4470588235294118, '
                            '0.6901960784313725), (0.8666666666666667, '
                            '0.5176470588235295, 0.3215686274509804), '
                            '(0.3333333333333333, 0.6588235294117647, '
                            '0.40784313725490196), (0.7686274509803922, '
                            '0.3058823529411765, 0.3215686274509804), '
                            '(0.5058823529411764, 0.4470588235294118, '
                            '0.7019607843137254), (0.5764705882352941, '
                            '0.47058823529411764, 0.3764705882352941), '
                            '(0.8549019607843137, 0.5450980392156862, '
                            '0.7647058823529411), (0.5490196078431373, '
                            '0.5490196078431373, 0.5490196078431373), (0.8, '
                            '0.7254901960784313, 0.4549019607843137), '
                            '(0.39215686274509803, 0.7098039215686275, '
                            '0.803921568627451)])',
         'axes.titlesize': 'original: large, new: 12.0',
         'font.sans-serif': "original: ['DejaVu Sans', 'Bitstream Vera Sans', "
                            "'Computer Modern Sans Serif', 'Lucida Grande', 'Verdana', "
                            "'Geneva', 'Lucid', 'Arial', 'Helvetica', 'Avant Garde', "
                            "'sans-serif'], new: ['Arial', 'DejaVu Sans', 'Liberation "
                            "Sans', 'Bitstream Vera Sans', 'sans-serif']",
         'font.size': 'original: 10.0, new: 12.0',
         'grid.color': 'original: #b0b0b0, new: .8',
         'grid.linewidth': 'original: 0.8, new: 1.0',
         'image.cmap': 'original: viridis, new: rocket',
         'legend.fontsize': 'original: medium, new: 11.0',
         'lines.solid_capstyle': 'original: projecting, new: round',
         'mathtext.bf': 'original: sans:bold, new: Bitstream Vera Sans:bold',
         'mathtext.fontset': 'original: dejavusans, new: custom',
         'mathtext.it': 'original: sans:italic, new: Bitstream Vera Sans:italic',
         'mathtext.rm': 'original: sans, new: Bitstream Vera Sans',
         'patch.edgecolor': 'original: black, new: w',
         'patch.facecolor': 'original: C0, new: (0.2980392156862745, '
                            '0.4470588235294118, 0.6901960784313725)',
         'patch.force_edgecolor': 'original: False, new: True',
         'text.color': 'original: black, new: .15',
         'text.usetex': 'original: False, new: True',
         'xtick.color': 'original: black, new: .15',
         'xtick.direction': 'original: out, new: in',
         'xtick.labelsize': 'original: medium, new: 11.0',
         'xtick.major.size': 'original: 3.5, new: 6.0',
         'xtick.major.width': 'original: 0.8, new: 1.25',
         'xtick.minor.size': 'original: 2.0, new: 4.0',
         'xtick.minor.top': 'original: True, new: False',
         'xtick.minor.visible': 'original: False, new: True',
         'xtick.minor.width': 'original: 0.6, new: 1.0',
         'ytick.color': 'original: black, new: .15',
         'ytick.direction': 'original: out, new: in',
         'ytick.labelsize': 'original: medium, new: 11.0',
         'ytick.major.size': 'original: 3.5, new: 6.0',
         'ytick.major.width': 'original: 0.8, new: 1.25',
         'ytick.minor.right': 'original: True, new: False',
         'ytick.minor.size': 'original: 2.0, new: 4.0',
         'ytick.minor.visible': 'original: False, new: True',
         'ytick.minor.width': 'original: 0.6, new: 1.0'}

    I implemented most of these below (although I left out a few color options).

    Args:
        None.
    Returns:
        None. The current matplotlib ``rcParams`` are modified.
    """
    # Color definitions from seaborn
    # NOTE: They need to be strings rather than raw floats.
    light_grey = ".8"
    # NOTE: I elect not to label with dark grey instead of black. It's not clear to me
    #       why that might be preferable here.

    # Setup the LaTeX preamble
    matplotlib.rcParams["text.latex.preamble"] = [
        # Enable AMS math package (for among other things, "\text")
        r"\usepackage{amsmath}",
        # Add fonts that will be used below. See the `mathtext` fonts set below for further info.
        r"\usepackage{sfmath}",
        # Ensure that existing values are included.
        matplotlib.rcParams["text.latex.preamble"],
    ]
    params = {
        # Enable latex
        "text.usetex": True,
        # Enable axis ticks (after they can be disabled by seaborn)
        "xtick.bottom": True,
        "ytick.left": True,
        # Make minor axis ticks visible (but only on left and bottom)
        "xtick.minor.visible": True,
        "ytick.minor.visible": True,
        "xtick.minor.top": False,
        "ytick.minor.right": False,
        # Ensure that axis ticks go inward instead of outward
        "xtick.direction": "in",
        "ytick.direction": "in",
        # Below, we set the LaTeX fonts to be the same fonts as those used in matplotlib.
        # For sans serif fonts in LaTeX (required for setting the fonts below), see: https://stackoverflow.com/a/11612347
        # To set the latex fonts to be the same as the normal matplotlib fonts, see: https://stackoverflow.com/a/27697390
        "mathtext.fontset": "custom",
        "mathtext.rm": "Bitstream Vera Sans",
        "mathtext.it": "Bitstream Vera Sans:italic",
        "mathtext.bf": "Bitstream Vera Sans:bold",
        ##### Extracted from seaborn
        # Plot axis underneath points
        "axes.axisbelow": True,
        # Modify label sizes.
        "axes.labelsize": 20.0,
        "axes.linewidth": 1.25,
        "axes.titlesize": 12.0,
        "font.size": 18.0,
        "legend.fontsize": 18.0,
        # Set the possible sans serif fonts. These are the ones made available in seaborn.
        "font.sans-serif": ["Arial", "DejaVu Sans", "Liberation " "Sans",
                            "Bitstream Vera Sans", "sans-serif"],
        # Make the grid lines light grey and slightly larger.
        "grid.color": light_grey,
        "grid.linewidth": 1.0,
        # End a line in a rounded style.
        "lines.solid_capstyle": "round",
        # This will disable lines connecting data points.
        # NOTE: This is disabled because if you forget to set the marker, then nothing will show up,
        #       which is a very frustrating user experience. Better to instead just disable it for a
        #       given plot.
        #"lines.linestyle": "none",
        # Set the edge color to white.
        "patch.edgecolor": "none",
        # Apparently this has to be enabled just for setting the edge color to be possible.
        "patch.force_edgecolor": True,
        # Tick label size.
        "xtick.labelsize": 20.0,
        "ytick.labelsize": 20.0,
        # Major tick settings
        "xtick.major.size": 6.0,
        "ytick.major.size": 6.0,
        "xtick.major.width": 1.25,
        "ytick.major.width": 1.25,
        # Minor tick settings
        "xtick.minor.size": 4.0,
        "ytick.minor.size": 4.0,
        "xtick.minor.width": 1.0,
        "ytick.minor.width": 1.0,
    }

    # Apply the updated settings.
    matplotlib.rcParams.update(params)




#!/usr/bin/env python

""" Physics Analysis Core for Heavy-Ions package, known as Pachyderm.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

__all__ = [
    "generic_class",
    "generic_config",
    "histogram",
    "plot",
    "projectors",
    "utils",
]

# Provide easy access to the version
# __version__ is the version string, while version_info is a tuple with an entry per point in the verion
from pachyderm.version import __version__   # noqa
from pachyderm.version import version_info  # noqa



#!/usr/bin/env python

""" Logging related fixtures to aid testing.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@yale.edu>, Yale University
"""

import logging
import pytest

# Set logging level as a global variable to simplify configuration.
# This is not ideal, but fine for simple tests.
logging_level = logging.DEBUG

@pytest.fixture
def logging_mixin(caplog):
    """ Logging mixin to capture logging messages from modules.

    It logs at the debug level, which is probably most useful for when a test fails.
    """
    caplog.set_level(logging_level)




#!/usr/bin/env python

""" Histogram related classes and functionality.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

from dataclasses import dataclass, field
import logging
import numpy as np
from types import TracebackType
from typing import Any, cast, ContextManager, Dict, Optional, Tuple, Type, TypeVar, Union

from pachyderm.typing_helpers import Axis, Hist, TFile

# Setup logger
logger = logging.getLogger(__name__)

_T_ContextManager = TypeVar("_T_ContextManager")

class RootOpen(ContextManager[_T_ContextManager]):
    """ Very simple helper to open root files. """
    def __init__(self, filename: str, mode: str = "read"):
        import ROOT
        self.filename = filename
        self.mode = mode
        self.f = ROOT.TFile.Open(self.filename, self.mode)

    def __enter__(self) -> TFile:
        if not self.f or self.f.IsZombie():
            raise IOError(f"Failed to open ROOT file '{self.filename}'.")
        return self.f

    def __exit__(self, execption_type: Optional[Type[BaseException]],
                 exception_value: Optional[BaseException], traceback: Optional[TracebackType]) -> None:
        """ We want to pass on all raised exceptions, but ensure that the file is always closed. """
        # The typing information is from here:
        # https://github.com/python/mypy/blob/master/docs/source/protocols.rst#context-manager-protocols

        # Pass on all of the exceptions, but make sure that the file is closed.
        # NOTE: The file isn't always valid because one has to deal with ROOT,
        #       so we have to explicitly check that is is valid before continuing.
        if self.f:
            self.f.Close()

        # We don't return anything because we always want the exceptions to continue
        # to be raised.

def get_histograms_in_file(filename: str) -> Dict[str, Any]:
    """ Helper function which gets all histograms in a file.

    Args:
        filename: Filename of the ROOT file containing the list.
    Returns:
        Contains hists with keys as their names. Lists are recursively added, mirroring
            the structure under which the hists were stored.
    """
    return get_histograms_in_list(filename = filename)

def get_histograms_in_list(filename: str, list_name: Optional[str] = None) -> Dict[str, Any]:
    """ Get histograms from the file and make them available in a dict.

    Lists are recursively explored, with all lists converted to dictionaries, such that the return
    dictionaries which only contains hists and dictionaries of hists (ie there are no ROOT ``TCollection``
    derived objects).

    Args:
        filename: Filename of the ROOT file containing the list.
        list_name: Name of the list to retrieve.
    Returns:
        Contains hists with keys as their names. Lists are recursively added, mirroring
            the structure under which the hists were stored.
    Raises:
        ValueError: If the list could not be found in the given file.
    """
    hists: Dict[str, Any] = {}
    with RootOpen(filename = filename, mode = "READ") as fIn:
        if list_name is not None:
            hist_list = fIn.Get(list_name)
        else:
            hist_list = [obj.ReadObj() for obj in fIn.GetListOfKeys()]

        if not hist_list:
            fIn.ls()
            # Closing this file appears (but is not entirely confirmed) to be extremely important! Otherwise,
            # the memory will leak, leading to ROOT memory issues!
            fIn.Close()
            raise ValueError(f"Could not find list with name \"{list_name}\". Possible names are listed above.")

        # Retrieve objects in the hist list
        for obj in hist_list:
            _retrieve_object(hists, obj)

    return hists

def _retrieve_object(output_dict: Dict[str, Any], obj: Any) -> None:
    """ Function to recursively retrieve histograms from a list in a ROOT file.

    ``SetDirectory(True)`` is applied to TH1 derived hists and python is explicitly given
    ownership of the retrieved objects.

    Args:
        output_dict (dict): Dict under which hists should be stored.
        obj (ROOT.TObject derived): Object(s) to be stored. If it is a collection,
            it will be recursed through.
    Returns:
        None: Changes in the dict are reflected in the output_dict which was passed.
    """
    import ROOT

    # Store TH1 or THn
    if isinstance(obj, ROOT.TH1) or isinstance(obj, ROOT.THnBase):
        # Ensure that it is not lost after the file is closed
        # Only works for TH1
        if isinstance(obj, ROOT.TH1):
            obj.SetDirectory(0)

        # Explicitly note that python owns the object
        # From more on memory management with ROOT and python, see:
        # https://root.cern.ch/root/html/guides/users-guide/PythonRuby.html#memory-handling
        ROOT.SetOwnership(obj, False)

        # Store the object
        output_dict[obj.GetName()] = obj

    # Recurse over lists
    if isinstance(obj, ROOT.TCollection):
        # Keeping it in order simply makes it easier to follow
        output_dict[obj.GetName()] = {}
        # Iterate over the objects in the collection and recursively store them
        for obj_temp in list(obj):
            _retrieve_object(output_dict[obj.GetName()], obj_temp)

# Typing helpers
_T = TypeVar("_T", bound = "Histogram1D")

@dataclass
class Histogram1D:
    """ Contains histogram data.

    Note:
        Underflow and overflow bins are excluded!

    Args:
        bin_edges (np.ndarray): The histogram bin edges.
        y (np.ndarray): The histogram bin values.
        errors_squared (np.ndarray): The bin sum weight squared errors.

    Attributes:
        x (np.ndarray): The bin centers.
        y (np.ndarray): The bin values.
        bin_edges (np.ndarray): The bin edges.
        errors (np.ndarray): The bin errors.
        errors_squared (np.ndarray): The bin sum weight squared errors.
        metadata (dict): Any additional metadata that should be stored with the histogram. Keys are expected to be
            strings, while the values can be anything. For example, could contain systematic errors, etc.
    """
    bin_edges: np.ndarray
    y: np.ndarray
    errors_squared: np.ndarray
    metadata: Dict[str, Any] = field(default_factory = dict)

    @property
    def errors(self) -> np.ndarray:
        return np.sqrt(self.errors_squared)

    @property
    def bin_widths(self) -> np.ndarray:
        """ Bin widths calculated from the bin edges.

        Returns:
            Array of the bin widths.
        """
        return self.bin_edges[1:] - self.bin_edges[:-1]

    @property
    def x(self) -> np.ndarray:
        """ The histogram bin centers (``x``).

        This property caches the x value so we don't have to calculate it every time.

        Args:
            None
        Returns:
            Array of center of bins.
        """
        try:
            return self._x
        except AttributeError:
            half_bin_widths = self.bin_widths / 2
            x = self.bin_edges[:-1] + half_bin_widths
            self._x: np.ndarray = x

        return self._x

    def find_bin(self, value: float) -> int:
        """ Find the bin corresponding to the specified value.

        For further information, see ``find_bin(...)`` in this module.

        Note:
            Bins are 0-indexed here, while in ROOT they are 1-indexed.

        Args:
            value: Value for which we want want the corresponding bin.
        Returns:
            Bin corresponding to the value.
        """
        return find_bin(self.bin_edges, value)

    def copy(self: _T) -> _T:
        """ Copies the object.

        In principle, this should be the same as ``copy.deepcopy(...)``, at least when this was written in
        Feb 2019. But ``deepcopy(...)`` often seems to have very bad performance (and perhaps does additional
        implicit copying), so we copy these numpy arrays by hand.
        """
        # We want to copy bin_edges, y, and errors_squared, but not anything else. Namely, we skip _x here.
        # In principle, it wouldn't really be a problem to copy, but there may be other "_" fields that we
        # want to skip later, so we do the right thing now.
        kwargs = {k: np.array(v, copy = True) for k, v in vars(self).items() if not k.startswith("_")}
        return type(self)(**kwargs)

    def counts_in_interval(self,
                           min_value: Optional[float] = None, max_value: Optional[float] = None,
                           min_bin: Optional[int] = None, max_bin: Optional[int] = None) -> Tuple[float, float]:
        """ Count the number of counts within bins in an interval.

        Note:
            The integration limits could be described as inclusive. This matches the ROOT convention.
            See ``histogram1D._integral(...)`` for further details on how these limits are determined.

        Note:
            The arguments can be mixed (ie. a min bin and a max value), so be careful!

        Args:
            min_value: Minimum value for the integral (we will find the bin which contains this value).
            max_value: Maximum value for the integral (we will find the bin which contains this value).
            min_bin: Minimum bin for the integral.
            max_bin: Maximum bin for the integral.
        Returns:
            (value, error): Integral value, error
        """
        return self._integral(
            min_value = min_value, max_value = max_value,
            min_bin = min_bin, max_bin = max_bin,
            multiply_by_bin_width = False,
        )

    def integral(self,
                 min_value: Optional[float] = None, max_value: Optional[float] = None,
                 min_bin: Optional[int] = None, max_bin: Optional[int] = None) -> Tuple[float, float]:
        """ Integrate the histogram over the given range.

        Note:
            The integration limits could be described as inclusive. This matches the ROOT convention.
            See ``histogram1D._integral(...)`` for further details on how these limits are determined.

        Note:
            The arguments can be mixed (ie. a min bin and a max value), so be careful!

        Args:
            min_value: Minimum value for the integral (we will find the bin which contains this value).
            max_value: Maximum value for the integral (we will find the bin which contains this value).
            min_bin: Minimum bin for the integral.
            max_bin: Maximum bin for the integral.
        Returns:
            (value, error): Integral value, error
        """
        return self._integral(
            min_value = min_value, max_value = max_value,
            min_bin = min_bin, max_bin = max_bin,
            multiply_by_bin_width = True,
        )

    def _integral(self,
                  min_value: Optional[float] = None, max_value: Optional[float] = None,
                  min_bin: Optional[int] = None, max_bin: Optional[int] = None,
                  multiply_by_bin_width: bool = False) -> Tuple[float, float]:
        """ Integrate the histogram over the specified range.

        This function provides the underlying implementation of the integral, giving the option to multiply
        by the bin with (in which case, one gets the integral), or not (in which case, one gets the number
        of counts in the range).

        Note:
            Limits of the integral could be described as inclusive. To understand this, consider an example
            where the bin edges are ``[0, 1, 2, 5]``, and we request value limits of ``(1.2, 3.6)``. The limits
            correspond to bins ``(1, 2)``, and therefore the integral will include the values from both bins 1 and 2.
            This matches the ROOT convention, and means that if a user wants the counts in only one bin, they
            should set the upper min and max bins to the same bin.

        Args:
            min_value: Minimum value for the integral (we will find the bin which contains this value).
            max_value: Maximum value for the integral (we will find the bin which contains this value).
            min_bin: Minimum bin for the integral.
            max_bin: Maximum bin for the integral.
            multiply_by_bin_width: If true, we will multiply each value by the bin width. The should be done
                for integrals, but not for counting values in an interval.
        Returns:
            (value, error): Integral value, error
        """
        # Validate arguments
        # Specified both values and bins, which is invalid.
        if min_value is not None and min_bin is not None:
            raise ValueError("Specified both min value and min bin. Only specify one.")
        if max_value is not None and max_bin is not None:
            raise ValueError("Specified both max value and max bin. Only specify one.")

        # Determine the bins from the values
        if min_value is not None:
            min_bin = self.find_bin(min_value)
        if max_value is not None:
            max_bin = self.find_bin(max_value)

        # Help out mypy.
        assert min_bin is not None
        assert max_bin is not None

        # Final validation to ensure that the bins properly ordered, with the min <= max.
        # NOTE: It is valid for the bins to be equal. In that case, we only take values from that single bin.
        if min_bin > max_bin:
            raise ValueError(
                f"Passed min_bin {min_bin} which is greater than the max_bin {max_bin}. The min bin must be smaller."
            )

        # Provide the opportunity to scale by bin width
        widths = np.ones(len(self.y))
        if multiply_by_bin_width:
            widths = self.bin_widths

        # Integrate by summing up all of the bins and the errors.
        # Perform the integral.
        # NOTE: We set the upper limits to + 1 from the found value because we want to include the bin
        #       where the upper limit resides. This matches the ROOT convention. Practically, this means
        #       that if the user wants to integrate over 1 bin, then the min bin and max bin should be the same.
        logger.debug(f"Integrating from {min_bin} - {max_bin + 1}")
        value = np.sum(self.y[min_bin:max_bin + 1] * widths[min_bin:max_bin + 1])
        error_squared = np.sum(self.errors_squared[min_bin:max_bin + 1] * widths[min_bin:max_bin + 1] ** 2)

        # We explicitly cast the final result to float to ensure that it doesn't cause any problems
        # with saving the final values to YAML.
        return float(value), float(np.sqrt(error_squared))

    def __add__(self: _T, other: _T) -> _T:
        """ Handles ``a = b + c.`` """
        new = self.copy()
        new += other
        return new

    def __radd__(self: _T, other: _T) -> _T:
        """ For use with sum(...). """
        if other == 0:
            return self
        else:
            return self + other

    def __iadd__(self: _T, other: _T) -> _T:
        """ Handles ``a += b``. """
        if not np.allclose(self.bin_edges, other.bin_edges):
            raise TypeError(
                f"Binning is different for given histograms."
                f"len(self): {len(self.bin_edges)}, len(other): {len(other.bin_edges)}."
                f"Cannot add!"
            )
        self.y += other.y
        self.errors_squared += other.errors_squared
        return self

    def __sub__(self: _T, other: _T) -> _T:
        """ Handles ``a = b - c``. """
        new = self.copy()
        new -= other
        return new

    def __isub__(self: _T, other: _T) -> _T:
        """ Handles ``a += b``. """
        if not np.allclose(self.bin_edges, other.bin_edges):
            raise TypeError(
                f"Binning is different for given histograms."
                f"len(self): {len(self.bin_edges)}, len(other): {len(other.bin_edges)}."
                f"Cannot subtract!"
            )
        self.y -= other.y
        self.errors_squared += other.errors_squared
        return self

    def __mul__(self: _T, other: _T) -> _T:
        """ Handles ``a = b * c``. """
        new = self.copy()
        new *= other
        return new

    def __imul__(self: _T, other: _T) -> _T:
        """ Handles ``a *= b``. """
        if not np.allclose(self.bin_edges, other.bin_edges):
            raise TypeError(
                f"Binning is different for given histograms."
                f"len(self): {len(self.bin_edges)}, len(other): {len(other.bin_edges)}."
                f"Cannot multiply!"
            )
        # NOTE: We need to calculate the errors_squared first because the depend on the existing y values
        # Errors are from ROOT::TH1::Multiply(const TH1 *h1)
        # NOTE: This is just error propagation, simplified with a = b * c!
        self.errors_squared = self.errors_squared * other.y ** 2 + other.errors_squared * self.y ** 2
        self.y *= other.y
        return self

    def __truediv__(self: _T, other: _T) -> _T:
        """ Handles ``a = b / c``. """
        new = self.copy()
        new /= other
        return new

    def __itruediv__(self: _T, other: _T) -> _T:
        """ Handles ``a /= b``. """
        if not np.allclose(self.bin_edges, other.bin_edges):
            raise TypeError(
                f"Binning is different for given histograms."
                f"len(self): {len(self.bin_edges)}, len(other): {len(other.bin_edges)}."
                f"Cannot divide!"
            )
        # Errors are from ROOT::TH1::Divide(const TH1 *h1)
        # NOTE: This is just error propagation, simplified with the a = b / c!
        # NOTE: We need to calculate the errors_squared first before setting y because the errors depend on
        #       the existing y values
        errors_squared_numerator = self.errors_squared * other.y ** 2 + other.errors_squared * self.y ** 2
        errors_squared_denominator = other.y ** 4
        # NOTE: We have to be a bit clever when we divide to avoid dividing by bins with 0 entries. The
        #       approach taken here basically replaces any divide by 0s with a 0 in the output hist.
        #       For more info, see: https://stackoverflow.com/a/37977222
        self.errors_squared = np.divide(
            errors_squared_numerator, errors_squared_denominator,
            out = np.zeros_like(errors_squared_numerator), where = errors_squared_denominator != 0,
        )
        self.y = np.divide(self.y, other.y, out = np.zeros_like(self.y), where = other.y != 0)
        return self

    def __eq__(self, other: Any) -> bool:
        """ Check for equality. """
        attributes = [k for k in vars(self) if not k.startswith("_")]
        other_attributes = [k for k in vars(other) if not k.startswith("_")]

        # As a beginning check, they must have the same attributes available.
        if attributes != other_attributes:
            return False

        # All attributes are np arrays, so we compare the arrays using ``np.allclose``
        # NOTE: allclose can't handle the metadata dict, so we skip it here and check
        #       it explicitly below.
        agreement = [np.allclose(getattr(self, a), getattr(other, a)) for a in attributes if a != "metadata"]
        # Check metadata
        metadata_agree = (self.metadata == other.metadata)
        # All arrays and the metadata must agree.
        return all(agreement) and metadata_agree

    @staticmethod
    def _from_uproot(hist: Any) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """ Convert a uproot histogram to a set of array for creating a Histogram.

        Note:
            Underflow and overflow bins are excluded!

        Args:
            hist (uproot.hist.TH1*): Input histogram.
        Returns:
            tuple: (x, y, errors) where x is the bin centers, y is the bin values, and
                errors are the sumw2 bin errors.
        """
        # This excludes underflow and overflow
        (y, bin_edges) = hist.numpy()

        # Also retrieve errors from sumw2.
        # If more sophistication is needed, we can modify this to follow the approach to
        # calculating bin errors from TH1::GetBinError()
        errors = hist.variances

        return (bin_edges, y, errors)

    @staticmethod
    def _from_th1(hist: Hist) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """ Convert a TH1 histogram to a Histogram.

        Note:
            Underflow and overflow bins are excluded!

        Args:
            hist (ROOT.TH1): Input histogram.
        Returns:
            tuple: (x, y, errors) where x is the bin centers, y is the bin values, and
                errors are the sumw2 bin errors.
        """
        # Enable sumw2 if it's not already calculated
        if hist.GetSumw2N() == 0:
            hist.Sumw2(True)

        # Don't include overflow
        bin_edges = get_bin_edges_from_axis(hist.GetXaxis())
        # NOTE: The y value and bin error are stored with the hist, not the axis.
        y = np.array([hist.GetBinContent(i) for i in range(1, hist.GetXaxis().GetNbins() + 1)])
        errors = np.array(hist.GetSumw2())
        # Exclude the under/overflow bins
        errors = errors[1:-1]

        return (bin_edges, y, errors)

    @classmethod
    def from_existing_hist(cls: Type[_T], hist: Union[Hist, Any]) -> _T:
        """ Convert an existing histogram.

        Note:
            Underflow and overflow bins are excluded! Bins are assumed to be fixed
            size.

        Args:
            hist (uproot.rootio.TH1* or ROOT.TH1): Histogram to be converted.
        Returns:
            Histogram: Dataclass with x, y, and errors
        """
        try:
            # Convert a histogram containing object -> TH1 or uproot hist.
            # It goes "HistogramContainer".hist -> TH1 or uproot hist
            #logger.debug("Converting HistogramContainer to standard hist")
            hist = hist.hist
        except AttributeError:
            # Just use the existing histogram
            pass
        # "values" is a proxy for if we have an uproot hist.
        logger.debug(f"{hist}, {type(hist)}")
        if hasattr(hist, "values"):
            (bin_edges, y, errors_squared) = cls._from_uproot(hist)
        else:
            # Handle traditional ROOT hists
            (bin_edges, y, errors_squared) = cls._from_th1(hist)

        return cls(bin_edges = bin_edges, y = y, errors_squared = errors_squared)

def get_array_from_hist2D(hist: Hist, set_zero_to_NaN: bool = True, return_bin_edges: bool = False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """ Extract x, y, and bin values from a 2D ROOT histogram.

    Converts the histogram into a numpy array, and suitably processes it for a surface plot
    by removing 0s (which can cause problems when taking logs), and returning a set of (x, y) mesh
    values utilziing either the bin edges or bin centers.

    Note:
        This is a different format than the 1D version!

    Args:
        hist (ROOT.TH2): Histogram to be converted.
        set_zero_to_NaN: If true, set 0 in the array to NaN. Useful with matplotlib so that it will
            ignore the values when plotting. See comments in this function for more details. Default: True.
        return_bin_edges: Return x and y using bin edges instead of bin centers.
    Returns:
        Contains (x values, y values, numpy array of hist data) where (x, y) are values on a
            grid (from np.meshgrid) using the selected bin values.
    """
    # Process the hist into a suitable state
    # NOTE: The shape specific can be somewhat confusing (ie. I would naviely expected to specify the x first.)
    # This says that the ``GetYaxis().GetNbins()`` number of rows and ``GetXaxis().GetNbins()`` number of columns.
    shape = (hist.GetYaxis().GetNbins(), hist.GetXaxis().GetNbins())
    # To keep consistency with the root_numpy 2D hist format, we transpose the final result
    # This format has x values as columns.
    hist_array = np.array([hist.GetBinContent(x) for x in range(1, hist.GetNcells()) if not hist.IsBinUnderflow(x) and not hist.IsBinOverflow(x)])
    # The hist_array was linear, so we need to shape it into our expected 2D values.
    hist_array = hist_array.reshape(shape)
    # Transpose the array to better match expectations
    # In particular, by transposing the array, it means that ``thist_array[1][0]`` gives the 2nd x
    # value (x_index = 1) and the 1st y value (y_index = 1). This is as we would expect. This is also
    # the same convention as used by root_numpy
    hist_array = hist_array.T
    # Set all 0s to nan to get similar behavior to ROOT. In ROOT, it will basically ignore 0s. This is
    # especially important for log plots. Matplotlib doesn't handle 0s as well, since it attempts to
    # plot them and then will throw exceptions when the log is taken.
    # By setting to nan, matplotlib basically ignores them similar to ROOT
    # NOTE: This requires a few special functions later which ignore nan when calculating min and max.
    if set_zero_to_NaN:
        hist_array[hist_array == 0] = np.nan

    if return_bin_edges:
        # Bin edges
        x_bin_edges = get_bin_edges_from_axis(hist.GetXaxis())
        y_bin_edges = get_bin_edges_from_axis(hist.GetYaxis())

        # NOTE: The addition of epsilon to the max is extremely important! Otherwise, the x and y
        #       ranges will be one bin short since ``arange`` is not inclusive. This could also be resolved
        #       by using ``linspace``, but I think this approach is perfectly fine.
        # NOTE: This epsilon is smaller than the one in ``utils`` because we are sometimes dealing
        #       with small times (~ns). The other value is larger because (I seem to recall) that
        #       smaller values didn't always place nice with ROOT, but it is fine here, since we're
        #       working with numpy.
        # NOTE: This should be identical to taking the min and max of the axis using
        #       ``TAxis.GetXmin()`` and ``TAxis.GetXmax()``, but I prefer this approach.
        epsilon = 1e-9
        x_range = np.arange(
            np.amin(x_bin_edges),
            np.amax(x_bin_edges) + epsilon,
            hist.GetXaxis().GetBinWidth(1)
        )
        y_range = np.arange(
            np.amin(y_bin_edges),
            np.amax(y_bin_edges) + epsilon,
            hist.GetYaxis().GetBinWidth(1)
        )
    else:
        # We want an array of bin centers
        x_range = np.array([hist.GetXaxis().GetBinCenter(i) for i in range(1, hist.GetXaxis().GetNbins() + 1)])
        y_range = np.array([hist.GetYaxis().GetBinCenter(i) for i in range(1, hist.GetYaxis().GetNbins() + 1)])

    X, Y = np.meshgrid(x_range, y_range)

    return (X, Y, hist_array)

def get_bin_edges_from_axis(axis: Axis) -> np.ndarray:
    """ Get bin edges from a ROOT hist axis.

    Note:
        Doesn't include over- or underflow bins!

    Args:
        axis (ROOT.TAxis): Axis from which the bin edges should be extracted.
    Returns:
        Array containing the bin edges.
    """
    # Don't include over- or underflow bins
    bins = range(1, axis.GetNbins() + 1)
    # Bin edges
    bin_edges = np.empty(len(bins) + 1)
    bin_edges[:-1] = [axis.GetBinLowEdge(i) for i in bins]
    bin_edges[-1] = axis.GetBinUpEdge(axis.GetNbins())

    return bin_edges

def find_bin(bin_edges: np.ndarray, value: float) -> int:
    """ Determine the index position where the value should be inserted.

    This is basically ``ROOT.TH1.FindBin(value)``, but it can used for any set of bin_edges.

    Note:
        Bins are 0-indexed here, while in ROOT they are 1-indexed.

    Args:
        bin_edges: Bin edges of the histogram.
        value: Value to find within those bin edges.
    Returns:
        Index of the bin where that value would reside in the histogram.
    """
    # This will return the index position where the value should be inserted.
    # This means that if we have the bin edges [0, 1, 2], and we pass value 1.5, it will return
    # index 2, but we want to return bin 1, so we subtract one from the result.
    # NOTE: By specifying that ``side = "right"``, it find values as arr[i] <= value < arr[i - 1],
    #       which matches the ROOT convention.
    return cast(
        int,
        np.searchsorted(bin_edges, value, side = "right") - 1
    )



#!/usr/bin/env python

""" Handle generic TH1 and THn projections.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import copy
import enum
import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from pachyderm import generic_class
from pachyderm.typing_helpers import Hist, Axis

# Setup logger
logger = logging.getLogger(__name__)

class TH1AxisType(enum.Enum):
    """ Map from (x,y,z) axis to the axis number.

    Other enumerations that refer to this enum should refer to the _values_ to ensure
    consistency in `.value` pointing to the axis value.
    """
    x_axis = 0
    y_axis = 1
    z_axis = 2

def hist_axis_func(axis_type: enum.Enum) -> Callable[[Hist], Axis]:
    """ Wrapper to retrieve the axis of a given histogram.

    This can be convenient outside of just projections, so it's made available in the API.

    Args:
        axis_type: The type of axis to retrieve.
    Returns:
        Callable to retrieve the specified axis when given a hist.
    """
    def axis_func(hist: Hist) -> Axis:
        """ Retrieve the axis associated with the ``HistAxisRange`` object for a given hist.

        Args:
            hist: Histogram from which the selected axis should be retrieved.
            axis_type: Enumeration corresponding to the axis to be restricted. The numerical
                value of the enum should be axis number (for a THnBase).
        Returns:
            ROOT.TAxis: The axis associated with the ``HistAxisRange`` object.
        """
        # Determine the axis_type value
        # Use try here instead of checking for a particular type to protect against type changes
        # (say in the enum)
        try:
            # Try to extract the value from an enum
            hist_axis_type = axis_type.value
        except AttributeError:
            # Seems that we received an int, so just use that value
            hist_axis_type = axis_type

        if hasattr(hist, "ProjectionND") and hasattr(hist, "Projection"):
            # THnBase defines ProjectionND and Projection, so we will use those as proxies.
            # Return the proper THn access
            #logger.debug(f"From hist: {hist}, hist_axis_type: {hist_axis_type}, axis: {hist.GetAxis(hist_axis_type.value)}")
            return hist.GetAxis(hist_axis_type)
        else:
            # If it's not a THn, then it must be a TH1 derived
            axis_function_map = {
                TH1AxisType.x_axis.value: hist.GetXaxis,
                TH1AxisType.y_axis.value: hist.GetYaxis,
                TH1AxisType.z_axis.value: hist.GetZaxis
            }

            # Retrieve the axis function and execute it. It is done separately to
            # clarify any possible errors.
            return_func = axis_function_map[hist_axis_type]
            return return_func()

    return axis_func

RangeMinMaxType = Union[float, Callable[[Any], float]]

class HistAxisRange(generic_class.EqualityMixin):
    """ Represents the restriction of a range of an axis of a histogram.

    An axis can be restricted by multiple ``HistAxisRange`` elements (although separate projections are
    needed to apply more than one. This would be accomplished with separate entries to the
    HistProjector.projection_dependent_cut_axes).

    NOTE:
        A single axis which has multiple ranges could be represented by multiple ``HistAxisRange`` objects!

    Args:
        axis_range_name (str): Name of the axis range. Usually some combination of the axis name and
            some sort of description of the range.
        axis_type (enum.Enum): Enumeration corresponding to the axis to be restricted. The numerical
            value of the enum should be axis number (for a THnBase).
        min_val (function): Minimum range value for the axis. Usually set via ``apply_func_to_find_bin()``.
        min_val (function): Maximum range value for the axis. Usually set via ``apply_func_to_find_bin()``.
    """
    def __init__(self, axis_range_name: str, axis_type: enum.Enum, min_val: RangeMinMaxType, max_val: RangeMinMaxType):
        self.name = axis_range_name
        self.axis_type = axis_type
        self.min_val = min_val
        self.max_val = max_val

    def __repr__(self) -> str:
        """ Representation of the object. """
        # The axis type is an enumeration of some type. In such a case, we want the repr to represent
        # it using the str method instead
        return "{}(name = {name!r}, axis_type = {axis_type}, min_val = {min_val!r}, max_val = {max_val!r})".format(self.__class__.__name__, **self.__dict__)

    def __str__(self) -> str:
        """ Print the elements of the object. """
        return "{}: name: {name}, axis_type: {axis_type}, min_val: {min_val}, max_val: {max_val}".format(self.__class__.__name__, **self.__dict__)

    @property
    def axis(self) -> Callable[[Any], Any]:
        """ Determine the axis to return based on the hist type. """
        axis_func = hist_axis_func(
            axis_type = self.axis_type
        )
        return axis_func

    def apply_range_set(self, hist: Hist) -> None:
        """ Apply the associated range set to the axis of a given hist.

        Note:
            The min and max values should be bins, not user ranges! For more, see the binning
            explanation in ``apply_func_to_find_bin(...)``.

        Args:
            hist: Histogram to which the axis range restriction should be applied.
        Returns:
            None. The range is set on the axis.
        """
        # Do individual assignments to clarify which particular value is causing an error here.
        axis = self.axis(hist)
        #logger.debug(f"axis: {axis}, axis(): {axis.GetName()}")
        # Help out mypy
        assert not isinstance(self.min_val, float)
        assert not isinstance(self.max_val, float)
        # Evaluate the functions to determine the values.
        min_val = self.min_val(axis)
        max_val = self.max_val(axis)
        # NOTE: Using SetRangeUser() here was a bug, since I've been passing bin values! In general,
        #       passing bin values is more flexible, but requires the values to be passed to
        #       ``apply_func_to_find_bin()`` to be shifted by some small epsilon to get the desired bin.
        self.axis(hist).SetRange(min_val, max_val)

    @staticmethod
    def apply_func_to_find_bin(
        func: Union[None, Callable[..., Union[float, int, Any]]],
        values: Optional[float] = None
    ) -> Callable[[Any], Union[float, int]]:
        """ Closure to determine the bin associated with a value on an axis.

        It can apply a function to an axis if necessary to determine the proper bin.  Otherwise,
        it can just return a stored value.

        Note:
            To properly determine the value, carefully note the information below. In many cases,
            such as when we want values [2, 5), the values need to be shifted by a small epsilon
            to retrieve the proper bin. This is done automatically in ``SetRangeUser()``.

            >>> hist = ROOT.TH1D("test", "test", 10, 0, 10)
            >>> x = 2, y = 5
            >>> hist.FindBin(x)
            2
            >>> hist.FindBin(x+epsilon)
            2
            >>> hist.FindBin(y)
            6
            >>> hist.FindBin(y-epsilon)
            5

            Note that the bin + epsilon on the lower bin is not strictly necessary, but it is
            used for consistency with the upper bound.

        Args:
            func (Callable): Function to apply to the histogram axis. If it is None, the value
                will be returned.
            values (int or float): Value to pass to the function. Default: None (in which case,
                it won't be passed).
        Returns:
            Function to be called with an axis to determine the desired bin on that axis.
        """
        def return_func(axis: Axis) -> Any:
            """ Apply the stored function and value to a given axis.

            Args:
                axis (TAxis or similar): Axis to which the function should be applied.
            Returns:
                any: The value returned by the function. Often a float or int, but not necessarily.
            """
            #logger.debug(f"func: {func}, values: {values}")
            if func:
                if values is not None:
                    return func(axis, values)
                else:
                    return func(axis)
            else:
                return values

        return return_func

class HistProjector:
    """ Handles generic ROOT ``THn`` and ``TH1`` projections.

    There are three types of cuts which can be specified:

    - ``additional_axis_cuts``: Axis cuts which do not change based on the projection axis.
    - ``projection_dependent_cut_axes``: Axis cuts which change based on the projection axis.
    - ``projection_axes``: Axes onto which the projection will be performed.

    For a full description of each type of cut and the necessary details, see their descriptions
    in the attributes.

    Note:
        The TH1 projections have not been tested as extensively as the ``THn`` projections.

    Note:
        ``input_key``, ``input_hist``, ``input_observable``, ``projection_name``, and ``output_hist`` are
        all reserved keys, such they will be overwritten by predefined information when passed to the
        various functions. Thus, they should be avoided by the user when storing projection information

    Args:
        observable_to_project_from: The observables which should be used to project from. The dict
            key is passed to ``projection_name(...)`` as ``input_key``.
        output_observable: Object or dict where the projected hist will be stored.
        projection_name_format: Format string to determine the projected hist name.
        output_attribute_name: Name of the attribute where which the single observable projection will
            be stored in the output_observable object. Must not be specified if projecting with multiple
            objects. Default: None.
        projection_information: Keyword arguments to be passed to ``projection_name(...)`` to determine
            the name of the projected histogram. Default: None.

    Attributes:
        single_observable_projection: True if the projector is only performing a single observable projection.
        output_attribute_name: Name of the attribute under which the single observable projection will be stored
            in the output_observable object.
        observable_to_project_from: The observable(s) which should be used to project from. The dict
            key is passed to ``projection_name(...)`` as ``input_key``.
        output_observable: Where the projected hist(s) will be stored. They will be stored under the dict
            key determined by ``output_key_name(...)``.
        projection_name_format: Format string to determine the projected hist name.
        projection_information: Keyword arguments to be passed to ``projection_name(...)`` to determine
            the name of the projected histogram.
        additional_axis_cuts (list): List of axis cuts which are neither projected nor depend on the axis
            being projected.
        projection_dependent_cut_axes (list): List of list of axis cuts which depend on the projected axis.
            For example, if we want to project non-continuous ranges of a non-projection axis (say, dEta
            when projecting dPhi). It is a list of list to allow for groups of cuts to be specified
            together if necessary.
        projection_axes (list): List of axes which should be projected.
    """
    def __init__(self,
                 observable_to_project_from: Union[Dict[str, Any], Hist, Any],
                 output_observable: Union[Dict[str, Any], Hist, Any],
                 projection_name_format: str,
                 output_attribute_name: Optional[str] = None,
                 projection_information: Optional[Dict[str, Any]] = None):
        # Determine whether we projecting multiple objects.
        # If not, store the attribute under which we are going to store the output.
        single_observable_projection = False
        if output_attribute_name is not None:
            single_observable_projection = True
        self.single_observable_projection = single_observable_projection
        self.output_attribute_name = output_attribute_name

        # Validate and assign input and output objects.
        if self.single_observable_projection and (isinstance(observable_to_project_from, dict) or isinstance(output_observable, dict)):
            raise ValueError("A dict of observables is not compatible with a single observable projection. Check your arguments.")
        self.observable_to_project_from = observable_to_project_from
        self.output_observable = output_observable

        # Output hist name format
        self.projection_name_format = projection_name_format

        # Additional projection information to help create names, input/output objects, etc
        # NOTE: See reserved keys enumerated above.
        if projection_information is None:
            projection_information = {}
        # Ensure that the dict is copied successfully
        self.projection_information = copy.deepcopy(projection_information)

        # Axes
        # Cuts for axes which are not projected
        self.additional_axis_cuts: List[HistAxisRange] = []
        # Axes cuts which depend on the projection axes
        # ie. If we want to change the range of the axis that we are projecting
        # For example, we may want to project an axis non-continuously (say, -1 - 0.5, 0.5 - 1)
        self.projection_dependent_cut_axes: List[List[HistAxisRange]] = []
        # Axes to actually project
        self.projection_axes: List[HistAxisRange] = []

    # Printing functions
    def __str__(self) -> str:
        """ Prints the properties of the projector.

        This will only show up properly when printed - otherwise the tabs and newlines won't be printed.
        """
        ret_val = f"{self.__class__.__name__}: Projection Information:\n"
        ret_val += f"\tprojection_name_format: \"{self.projection_name_format}\""
        ret_val += "\n\tprojection_information:\n"
        ret_val += "\n".join(["\t\t- " + str("Arg: ") + str(val) for arg, val in self.projection_information.items()])
        ret_val += "\n\tadditional_axis_cuts:\n"
        ret_val += "\n".join(["\t\t- " + str(axis) for axis in self.additional_axis_cuts])
        ret_val += "\n\tprojection_dependent_cut_axes:\n"
        ret_val += "\n".join(["\t\t- " + str([",".join(axis.name for axis in axis_list)]) for axis_list in self.projection_dependent_cut_axes])
        ret_val += "\n\tprojection_axes:\n"
        ret_val += "\n".join(["\t\t- " + str(axis) for axis in self.projection_axes])

        return ret_val

    def call_projection_function(self, hist: Hist) -> Hist:
        """ Calls the actual projection function for the hist.

        Args:
            hist: Histogram from which the projections should be performed.
        Returns:
            The projected histogram.
        """
        # Restrict projection axis ranges
        for axis in self.projection_axes:
            logger.debug(f"Apply projection axes hist range: {axis.name}")
            axis.apply_range_set(hist)

        projected_hist = None
        if hasattr(hist, "ProjectionND") and hasattr(hist, "Projection"):
            # THnBase defines ProjectionND and Projection, so we will use those as proxies.
            projected_hist = self._project_THn(hist = hist)
        elif hasattr(hist, "ProjectionZ") and hasattr(hist, "Project3D"):
            # TH3 defines ProjectionZ and Project3D, so we will use those as proxies.
            projected_hist = self._project_TH3(hist = hist)
        elif hasattr(hist, "ProjectionX") and hasattr(hist, "ProjectionY"):
            # TH2 defines ProjectionX and ProjectionY, so we will use those as proxies.
            projected_hist = self._project_TH2(hist = hist)
        else:
            raise TypeError(type(hist), f"Could not recognize hist {hist} of type {type(hist)}")

        # Cleanup restricted axes
        self.cleanup_cuts(hist, cut_axes = self.projection_axes)

        return projected_hist

    def _project_THn(self, hist: Hist) -> Any:
        """ Perform the actual THn -> THn or TH1 projection.

        This projection could be to 1D, 2D, 3D, or ND.

        Args:
            hist (ROOT.THnBase): Histogram from which the projections should be performed.
        Returns:
            ROOT.THnBase or ROOT.TH1: The projected histogram.
        """
        # THnBase projections args are given as a list of axes, followed by any possible options.
        projection_axes = [axis.axis_type.value for axis in self.projection_axes]

        # Handle ROOT THnBase quirk...
        # 2D projection are called as (y, x, options), so we should reverse the order so it performs
        # as expected
        if len(projection_axes) == 2:
            # Reverses in place
            projection_axes.reverse()

        # Test calculating errors
        # Add "E" to ensure that errors will be calculated
        args = projection_axes + ["E"]
        # Do the actual projection
        logger.debug(f"hist: {hist.GetName()} args: {args}")

        if len(projection_axes) > 3:
            # Project into a THnBase object.
            projected_hist = hist.ProjectionND(*args)
        else:
            # Project a TH1 derived object.
            projected_hist = hist.Projection(*args)

        return projected_hist

    def _project_TH3(self, hist: Hist) -> Any:
        """ Perform the actual TH3 -> TH1 projection.

        This projection could be to 1D or 2D.

        Args:
            hist (ROOT.TH3): Histogram from which the projections should be performed.
        Returns:
            ROOT.TH1: The projected histogram.
        """
        # Axis length validation
        if len(self.projection_axes) < 1 or len(self.projection_axes) > 2:
            raise ValueError(len(self.projection_axes), "Invalid number of axes")

        # Need to concatenate the names of the axes together
        projection_axis_name = ""
        for axis in self.projection_axes:
            # Determine the axis name based on the name of the axis type.
            # [:1] returns just the first letter. For example, we could get "xy" if the first axis as
            # x_axis and the second was y_axis.
            # NOTE: Careful. This depends on the name of the enumerated values!!! Since this isn't terribly
            #       safe, we then perform additional validation on the same to ensure that it is one of the
            #       expected axis names.
            proj_axis_name = axis.axis_type.name[:1]
            if proj_axis_name not in ["x", "y", "z"]:
                raise ValueError(f"Projection axis name {proj_axis_name} is not 'x', 'y', or 'z'. Please check your configuration.")
            projection_axis_name += proj_axis_name

        # Handle ROOT Project3D quirk...
        # 2D projection are called as (y, x, options), so we should reverse the order so it performs
        # as expected.
        # NOTE: This isn't well documented in TH3. It is instead described in THnBase.Projection(...)
        if len(self.projection_axes) == 2:
            # Reverse the axes
            projection_axis_name = projection_axis_name[::-1]

        # Do the actual projection
        logger.info(f"Projecting onto axes \"{projection_axis_name}\" from hist {hist.GetName()}")
        projected_hist = hist.Project3D(projection_axis_name)

        return projected_hist

    def _project_TH2(self, hist: Hist) -> Any:
        """ Perform the actual TH2 -> TH1 projection.

        This projection can only be to 1D.

        Args:
            hist (ROOT.TH2): Histogram from which the projections should be performed.
        Returns:
            ROOT.TH1: The projected histogram.
        """
        if len(self.projection_axes) != 1:
            raise ValueError(len(self.projection_axes), "Invalid number of axes")

        #logger.debug(f"self.projection_axes[0].axis: {self.projection_axes[0].axis}, axis range name: {self.projection_axes[0].name}, axis_type: {self.projection_axes[0].axis_type}")
        # NOTE: We cannot use TH3.ProjectionZ(...) because it has different semantics than ProjectionX
        #       and ProjectionY. In particular, it doesn't respect the axis limits of axis onto which it
        #       is projected.  So we have to separate the projection by histogram type as opposed to axis
        #       length.
        projection_func_map = {
            TH1AxisType.x_axis.value: hist.ProjectionX,
            TH1AxisType.y_axis.value: hist.ProjectionY
        }

        # Determine the axis_type value
        # Use try here instead of checking for a particular type to protect against type changes (say
        # in the enum)
        try:
            # Try to extract the value from an enum
            axis_type = self.projection_axes[0].axis_type.value
        except ValueError:
            # Seems that we received an int, so just use that value
            axis_type = self.axis_type  # type: ignore

        projection_func = projection_func_map[axis_type]

        # Do the actual projection
        logger.info(f"Projecting onto axis range {self.projection_axes[0].name} from hist {hist.GetName()}")
        projected_hist = projection_func()

        return projected_hist

    def _project_observable(self, input_key: str,
                            input_observable: Any,
                            get_hist_args: Optional[Dict[str, Any]] = None,
                            projection_name_args: Optional[Dict[str, Any]] = None,
                            **kwargs: Any) -> Hist:
        """ Perform a projection for a single observable.

        Note:
            All cuts on the original histograms will be reset when this function is completed.

        Args:
            input_key: Key to describe the input observable.
            input_observable: Observable to project from.
            get_hist_args: Arguments to pass to ``get_hist(...)``. Made available so the args can be cached
                to avoid a ``deepcopy`` when looping. Default: None. In this case, they will be retrieved
                automatically.
            projection_name_args: Arguments to pass to ``projection_name(...)``. Made available so the args
                can be cached to avoid a ``deepcopy`` when looping. Default: None. In this case, they will be
                retrieved automatically.
            kwargs: Additional named args to be passed to projection_name(...) and output_key_name(...).
        Returns:
            The projected histogram.
        """
        # Validation of other optional arguments.
        if get_hist_args is None:
            get_hist_args = copy.deepcopy(kwargs)
        if projection_name_args is None:
            projection_name_args = copy.deepcopy(kwargs)

        # Retrieve histogram
        # We update ``input_observable`` in ``get_hist_args`` every loop, so we don't have to worry
        # about passing the wrong observable.
        get_hist_args.update({"observable": input_observable})
        hist = self.get_hist(**get_hist_args)

        # Define projection name
        projection_name_args.update(self.projection_information)
        # In principle, we could have overwritten one of the kwargs, so we ensure with one of the other
        # updates, so we update it again to be certain.
        projection_name_args.update(kwargs)
        # Put the values included by default last to ensure nothing overwrites these values
        projection_name_args.update({
            "input_key": input_key,
            "input_observable": input_observable,
            "input_hist": hist
        })
        projection_name = self.projection_name(**projection_name_args)

        # First apply the cuts
        # Restricting the range with SetRange(User) works properly for both THn and TH1.
        logger.debug(f"hist: {hist}")
        for axis in self.additional_axis_cuts:
            logger.debug(f"Apply additional axis hist range: {axis.name}")
            axis.apply_range_set(hist)

        # We need to ensure that it isn't empty so at least one project occurs
        if self.projection_dependent_cut_axes == []:
            self.projection_dependent_cut_axes.append([])

        # Validate the projection dependent cut axes
        # It is invalid to have PDCA on the same axes as the projection axes.
        duplicated_axes = [
            PDCA
            for PA in self.projection_axes
            for PDCA_group in self.projection_dependent_cut_axes
            for PDCA in PDCA_group
            if PDCA.axis_type == PA.axis_type
        ]
        if duplicated_axes:
            raise ValueError(
                f"Axis {duplicated_axes} is in the projection axes and the projection dependent cut axes."
                " This configuration is not allowed, as the range in the PDCA will be overwritten by the projection axes!"
                " Please revise your configuration."
            )

        # Perform the projections
        hists = []
        for i, axes in enumerate(self.projection_dependent_cut_axes):
            # Projection dependent range set
            for axis in axes:
                logger.debug(f"Apply projection dependent hist range: {axis.name}")
                axis.apply_range_set(hist)

            # Do the projection
            projected_hist = self.call_projection_function(hist)
            projected_hist.SetName(f"{projection_name}_{i}")

            hists.append(projected_hist)

            # Cleanup projection dependent cuts (although they should be set again on the next
            # iteration of the loop)
            self.cleanup_cuts(hist, cut_axes = axes)

        # Cleanup the rest of the cuts
        self.cleanup_cuts(hist, cut_axes = self.additional_axis_cuts)

        # Combine all of the projections together
        output_hist = hists[0]
        for temp_hist in hists[1:]:
            output_hist.Add(temp_hist)

        # Final settings
        output_hist.SetName(projection_name)
        # Ensure that the hist doesn't get deleted by ROOT
        # A reference to the histogram within python may not be enough
        output_hist.SetDirectory(0)

        return output_hist, projection_name, projection_name_args

    def _project_single_observable(self, **kwargs: Any) -> Hist:
        """ Driver function for projecting and storing a single observable.

        Args:
            kwargs (dict): Additional named args to be passed to projection_name(...) and output_key_name(...)
        Returns:
            The projected histogram. The histogram is also stored in the output specified by ``output_observable``.
        """
        # Help out mypy
        assert isinstance(self.output_attribute_name, str)

        # Run the actual projection.
        output_hist, projection_name, projection_name_args, = self._project_observable(
            input_key = "single_observable",
            input_observable = self.observable_to_project_from,
            **kwargs,
        )
        # Store the output.
        output_hist_args = projection_name_args
        output_hist_args.update({
            "output_hist": output_hist,
            "projection_name": projection_name
        })

        # Store the final histogram.
        output_hist = self.output_hist(**output_hist_args)

        # Store the final output hist
        if not hasattr(self.output_observable, self.output_attribute_name):
            raise ValueError(f"Attempted to assign hist to non-existent attribute {self.output_attribute_name} of object {self.output_observable}. Check the attribute name!")
        # Actually store the histogram.
        setattr(self.output_observable, self.output_attribute_name, output_hist)

        # Return the observable
        return output_hist

    def _project_dict(self, **kwargs: Any) -> Dict[str, Hist]:
        """ Driver function for projecting and storing a dictionary of observables.

        Args:
            kwargs (dict): Additional named args to be passed to projection_name(...) and output_key_name(...)
        Returns:
            The projected histograms. The projected histograms are also stored in ``output_observable``.
        """
        # Setup function arguments with values which don't change per loop.
        get_hist_args = copy.deepcopy(kwargs)
        projection_name_args = copy.deepcopy(kwargs)
        for key, input_observable in self.observable_to_project_from.items():
            output_hist, projection_name, projection_name_args, = self._project_observable(
                input_key = key,
                input_observable = input_observable,
                get_hist_args = get_hist_args,
                projection_name_args = projection_name_args,
                **kwargs,
            )

            # Store the output observable
            output_hist_args = projection_name_args
            output_hist_args.update({
                "output_hist": output_hist,
                "projection_name": projection_name
            })
            output_key_name = self.output_key_name(**output_hist_args)
            self.output_observable[output_key_name] = self.output_hist(**output_hist_args)

        return self.output_observable

    def project(self, **kwargs: Any) -> Union[Hist, Dict[str, Hist]]:
        """ Perform the requested projection(s).

        Note:
            All cuts on the original histograms will be reset when this function is completed.

        Args:
            kwargs (dict): Additional named args to be passed to projection_name(...) and output_key_name(...)
        Returns:
            The projected histogram(s). The projected histograms are also stored in ``output_observable``.
        """
        if self.single_observable_projection:
            return self._project_single_observable(**kwargs)
        else:
            return self._project_dict(**kwargs)

    def cleanup_cuts(self, hist: Hist, cut_axes: Iterable[HistAxisRange]) -> None:
        """ Cleanup applied cuts by resetting the axis to the full range.

        Inspired by: https://github.com/matplo/rootutils/blob/master/python/2.7/THnSparseWrapper.py

        Args:
            hist: Histogram for which the axes should be reset.
            cut_axes: List of axis cuts, which correspond to axes that should be reset.
        """
        for axis in cut_axes:
            # According to the function TAxis::SetRange(first, last), the widest possible range is
            # (1, Nbins). Anything beyond that will be reset to (1, Nbins)
            axis.axis(hist).SetRange(1, axis.axis(hist).GetNbins())

    #############################
    # Functions to be overridden!
    #############################
    def projection_name(self, **kwargs: Dict[str, Any]) -> str:
        """ Define the projection name for this projector.

        Note:
            This function is just a basic placeholder and likely should be overridden.

        Args:
            kwargs: Projection information dict combined with additional arguments passed to the
                projection function.
        Returns:
            Projection name string formatted with the passed options. By default, it returns
                ``projection_name_format`` formatted with the arguments to this function.
        """
        return self.projection_name_format.format(**kwargs)

    def get_hist(self, observable: Any, **kwargs: Dict[str, Any]) -> Any:
        """ Get the histogram that may be stored in some object.

        This histogram is used to project from.

        Note:
            The output object could just be the raw ROOT histogram.

        Note:
            This function is just a basic placeholder and likely should be overridden.

        Args:
            observable (object): The input object. It could be a histogram or something more complex
            kwargs: Additional arguments passed to the projection function
        Return:
            ROOT.TH1 or ROOT.THnBase histogram which should be projected. By default, it returns the
                observable (input object).
        """
        return observable

    def output_key_name(self, input_key: str, output_hist: Hist, projection_name: str, **kwargs: str) -> str:
        """ Returns the key under which the output object should be stored.

        Note:
            This function is just a basic placeholder which returns the projection name
            and likely should be overridden.

        Args:
            input_key: Key of the input hist in the input dict
            output_hist: The output histogram
            projection_name: Projection name for the output histogram
            kwargs: Projection information dict combined with additional arguments passed to
                the projection function.
        Returns:
            Key under which the output object should be stored. By default, it returns the
                projection name.
        """
        return projection_name

    def output_hist(self, output_hist: Hist, input_observable: Any, **kwargs: Any) -> Union[Hist, Any]:
        """ Return an output object. It should store the ``output_hist``.

        Note:
            The output object could just be the raw histogram.

        Note:
            This function is just a basic placeholder which returns the given output object (a histogram)
            and likely should be overridden.

        Args:
            output_hist: The output histogram
            input_observable (object): The corresponding input object. It could be a histogram or something
                more complex.
            kwargs: Projection information dict combined with additional arguments passed to the
                projection function
        Return:
            The output object which should be stored in the output dict. By default, it returns the
                output hist.
        """
        return output_hist




#!/usr/bin/env python

""" Define the pachyderm package version in a single location.

Inspired by method 3 `here <https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version>`__,
as well as ``uproot`` (which more or less implements this pattern). Further discussion is available
`here <https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package>`__.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import re

# Provide the version as a string
__version__ = "1.9.6"
version = __version__
# As provide as a tuple, with each value as it's own entry.
# ie. "1.0" -> `("1", "0")`
version_info = tuple(re.split(r"[-\.]", __version__))

del re
